package ru.dbotthepony.kommons.collect

import java.util.BitSet

sealed class BitSetIterator(protected val set: BitSet, first: Int, private val last: Int) : IntIterator() {
	protected abstract fun findNext(from: Int): Int

	private var hasNext = false
	private var next = -1

	init {
		if (first < last) {
			val nextSet = findNext(first)

			if (nextSet < last && nextSet != -1) {
				hasNext = true
				next = nextSet
			}
		}
	}

	override fun nextInt(): Int {
		if (!hasNext)
			throw NoSuchElementException()

		val nextSet = findNext(next + 1)

		if (nextSet < last && nextSet != -1) {
			hasNext = true
			val value = next
			next = nextSet
			return value
		} else {
			hasNext = false
			return next
		}
	}

	override fun hasNext(): Boolean {
		return hasNext
	}

	class Set(set: BitSet, first: Int, last: Int) : BitSetIterator(set, first, last) {
		override fun findNext(from: Int): Int {
			return set.nextSetBit(from)
		}
	}

	class Clear(set: BitSet, first: Int, last: Int) : BitSetIterator(set, first, last) {
		override fun findNext(from: Int): Int {
			return set.nextClearBit(from)
		}
	}
}

/**
 * Iterates set bits in this [BitSet], from [first] (inclusive) up to [last] (exclusive)
 */
fun BitSet.iterateSetBits(first: Int, last: Int): IntIterator {
	return BitSetIterator.Set(this, first, last)
}

/**
 * Iterates clear bits in this [BitSet], from [first] (inclusive) up to [last] (exclusive)
 */
fun BitSet.iterateClearBits(first: Int, last: Int): IntIterator {
	return BitSetIterator.Clear(this, first, last)
}

/**
 * Iterates set bits in this [BitSet], from 0 (inclusive) up to [last] (exclusive)
 */
fun BitSet.iterateSetBits(last: Int): IntIterator {
	return BitSetIterator.Set(this, 0, last)
}

/**
 * Iterates clear bits in this [BitSet], from 0 (inclusive) up to [last] (exclusive)
 */
fun BitSet.iterateClearBits(last: Int): IntIterator {
	return BitSetIterator.Clear(this, 0, last)
}
