package ru.dbotthepony.kommons.math

class HSVColor(hue: Float, saturation: Float, value: Float) : Comparable<HSVColor> {
	val hue = (hue % 360f).let { if (it < 0f) it + 360f else it }
	val saturation = saturation.coerceIn(0f, 1f)
	val value = value.coerceIn(0f, 1f)

	override fun equals(other: Any?): Boolean {
		return other === this || other is HSVColor && other.hue == hue && other.saturation == saturation && other.value == value
	}

	override fun hashCode(): Int {
		return hue.hashCode() + saturation.hashCode() * 31 + value.hashCode() * 31 * 31
	}

	fun copy(hue: Float = this.hue, saturation: Float = this.saturation, value: Float = this.value): HSVColor {
		return HSVColor(hue, saturation, value)
	}

	operator fun component1() = hue
	operator fun component2() = saturation
	operator fun component3() = value

	fun toRGBA(alpha: Float = 1f): RGBAColor {
		val valueMin = (1f - saturation) * value
		val delta = (value - valueMin) * (hue % 60f) / 60f
		val valueInc = valueMin + delta
		val valueDec = value - delta

		return when ((hue / 60f).toInt()) {
			0 -> RGBAColor(value, valueInc, valueMin, alpha)
			1 -> RGBAColor(valueDec, value, valueMin, alpha)
			2 -> RGBAColor(valueMin, value, valueInc, alpha)
			3 -> RGBAColor(valueMin, valueDec, value, alpha)
			4 -> RGBAColor(valueInc, valueMin, value, alpha)
			5 -> RGBAColor(value, valueMin, valueDec, alpha)
			else -> throw IllegalStateException("whut")
		}
	}

	override fun compareTo(other: HSVColor): Int {
		return comparator.compare(this, other)
	}

	companion object {
		@JvmField val WHITE = HSVColor(0f, 1f, 1f)

		private val comparator = Comparator
			.comparing(HSVColor::hue)
			.thenComparing(HSVColor::saturation)
			.thenComparing(HSVColor::value)
	}
}
