package ru.dbotthepony.kommons.random

import java.util.random.RandomGenerator

/**
 * Classical LCG PRNG, utilizing [MULTIPLIER] and [INCREMENT] as constants.
 *
 * Internal state is presented as 64-bit integer, and high-order bits of internal state are
 * used to generate numbers.
 *
 * Generally shouldn't be used, and instead [PCG32Random] should be used, considering the latter has same memory
 * requirements but has way better statistical properties.
 */
open class LCG64Random(
	@JvmField
	protected var seed: Long
) : RandomGenerator {
	override fun nextLong(): Long {
		val a = nextInt().toLong()
		val b = nextInt().toLong() and 0xFFFFFFFFL
		return a.shl(32) or b
	}

	override fun nextInt(): Int {
		this.seed = MULTIPLIER * this.seed + INCREMENT
		return this.seed.ushr(32).toInt()
	}

	companion object {
		const val MULTIPLIER = 6364136223846793005
		const val INCREMENT = 1442695040888963407
	}
}
