package ru.dbotthepony.kommons.random

import java.util.random.RandomGenerator

/**
 * PCG XSH RR 32-bit (64-bit state) random generator, which has way better statistical properties
 * than plain [LCG64Random] generator
 */
open class PCG32Random(
	@JvmField
	protected var seed: Long
) : RandomGenerator {
	final override fun nextInt(): Int {
		var x = this.seed
		this.seed = LCG64Random.MULTIPLIER * this.seed + LCG64Random.INCREMENT

		val rot = x.ushr(59) and MASK
		x = x.xor(x.ushr(18))
		return x.ushr(27).toInt().rotateLeft(rot.toInt())
	}

	override fun nextLong(): Long {
		val a = nextInt().toLong()
		val b = nextInt().toLong() and 0xFFFFFFFFL
		return a.shl(32) or b
	}

	companion object {
		const val MASK = (1L shl 5) - 1L
	}
}
